# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Script to fix problems in python models generated by datamodel-code-generator.

This script does cleanup on python files generated by datamodel-codegenerator.
The script changes the Any's appropriately from the plural form of the needed
  class to the singular form using a manually curated lookup table.
It also finds all references in a class to classes it depends on, using regex.
It then reorders the classes so there are minimal forwards required.
This script is normally called by gen_oscal.py when models are generated.
"""
import logging
import operator
import re
import string

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler())

# these plurals need to be fixed in Dict even when non-optional
special_plurals = ['Users', 'Components']

plural_lut = {
    'Annotations': 'Annotation',
    'AssessmentSubjects': 'AssessmentSubject',
    'ByComponents': 'ByComponent',
    'Capabilities': 'Capability',
    'Components': 'SystemComponent',
    'Controls': 'Control',
    'Diagrams': 'Diagram',
    'EmailAddresses': 'EmailStr',
    'ImplementedComponents': 'ImplementedComponent',
    'IncorporatesComponents': 'IncorporatesComponent',
    'IncorporatesComponentDefnitions': 'IncorporatesComponentDefinition',
    'IncorporatesTargets': 'IncorporatesTarget',
    'InformationTypeIds': 'InformationTypeId',
    'InventoryItems': 'InventoryItem',
    'Objectives': 'Objective',
    'ParameterSettings': 'SetParameter',
    'ResponsibleParties': 'ResponsibleParty',
    'ResponsibleRoles': 'ResponsibleRole',
    'SetParameters': 'SetParameter',
    'Statements': 'Statement',
    'Tools': 'Tool',
    'Targets': 'DefinedTarget',
    'Users': 'SystemUser'
}

# Class names with this substring are due to confusion in oscal that will go away.
# For now delete the classes and convert the corresponding unions to conlist
singleton_suffixes = ['GroupItem', 'Item']
singleton_prefixes = ['Include', 'Exclude']
numbered_classes = ['Base64']
class_header = 'class '
license_header = (
    '# -*- mode:python; coding:utf-8 -*-\n'
    '# Copyright (c) 2020 IBM Corp. All rights reserved.\n'
    '#\n'
    '# Licensed under the Apache License, Version 2.0 (the "License");\n'
    '# you may not use this file except in compliance with the License.\n'
    '# You may obtain a copy of the License at\n'
    '#\n'
    '#     http://www.apache.org/licenses/LICENSE-2.0\n'
    '#\n'
    '# Unless required by applicable law or agreed to in writing, software\n'
    '# distributed under the License is distributed on an "AS IS" BASIS,\n'
    '# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.\n'
    '# See the License for the specific language governing permissions and\n'
    '# limitations under the License.\n'
)


class RelOrder():
    """Capture relative location of each class in list to its refs and deps."""

    def __init__(self, max_index):
        """Initialize with size of list being reordered."""
        self.latest_dep = 0
        self.earliest_ref = max_index


class ClassText():
    """Hold class text as named blocks with references to the added classes and capture its refs."""

    def __init__(self, first_line):
        """Construct with first line of class definition."""
        self.lines = [first_line.rstrip()]
        n = first_line.find('(')
        self.name = first_line[len(class_header):n]
        self.refs = set()
        self.full_refs = set()
        self.found_all_links = False
        self.is_self_ref = False

    def add_line(self, line):
        """Add new line to class text."""
        self.lines.append(line)

    def add_ref_if_good(self, ref_name):
        """Add refs after removing digits and/or singleton string."""
        if ref_name == 'Base64':
            self.refs.add(ref_name)
            return
        for suffix in singleton_suffixes:
            n = ref_name.find(suffix)
            if n > 0:
                ref_name = ref_name[:n]
                break
        for prefix in singleton_prefixes:
            n = ref_name.find(prefix)
            if n == 0 and ref_name != prefix:
                ref_name = ref_name[len(prefix):]
        self.refs.add(ref_name.rstrip(string.digits))

    def add_ref_pattern(self, p, line):
        """Add new class names found based on pattern."""
        new_refs = p.findall(line)
        if new_refs:
            for r in new_refs:
                if type(r) == tuple:
                    for s in r:
                        self.add_ref_if_good(s)
                else:
                    self.add_ref_if_good(r)

    @staticmethod
    def find_index(class_text_list, name):
        """Find index of class in list by name."""
        nclasses = len(class_text_list)
        for i in range(nclasses):
            if class_text_list[i].name == name:
                return i
        return -1

    def add_all_refs(self, line):
        """Find all refd class names found in line and add to references."""
        # find lone strings with no brackets
        p = re.compile(r'.*\:\s*([^\s\[\]]+).*')
        self.add_ref_pattern(p, line)
        # find objects in one or more bracket sets with possible first token and comma
        p = re.compile(r'.*\[(?:(.*),\s*)?((?:\[??[^\[]*?))\]')
        self.add_ref_pattern(p, line)
        return line

    def get_linked_refs(self, class_text_list, known_refs):
        """Follow all refs to find their refs."""
        if self.name in known_refs:
            return set()
        refs = self.full_refs
        if self.found_all_links:
            return refs
        new_refs = refs.copy()
        for r in refs:
            i = ClassText.find_index(class_text_list, r)
            new_refs = new_refs.union(class_text_list[i].get_linked_refs(class_text_list, refs))
        self.found_all_links = True
        return new_refs

    def find_direct_refs(self, class_names_list):
        """Find direct refs without recursion."""
        for ref in self.refs:
            if ref == self.name:
                self.is_self_ref = True
            if ref in class_names_list and not ref == self.name:
                self.full_refs.add(ref)
        if len(self.full_refs) == 0:
            self.found_all_links = True

    def find_full_refs(self, class_text_list):
        """Find full refs with recursion."""
        sub_refs = set()
        for ref in self.full_refs:
            # find the class named ref
            i = ClassText.find_index(class_text_list, ref)
            # accumulate all references associated with each ref in that class
            sub_refs = sub_refs.union(class_text_list[i].get_linked_refs(class_text_list, sub_refs))
        return self.full_refs.union(sub_refs)

    def find_order(self, class_text_list):
        """Find latest dep and earliest reference."""
        ro = RelOrder(len(class_text_list) - 1)
        # find first class that needs this class
        for i, ct in enumerate(class_text_list):
            if self.name in ct.full_refs:
                ro.earliest_ref = i
                break
        # find last class this one needs
        # make sure result is deterministic and does not depend on order from set
        sorted_ref_list = sorted(self.full_refs)
        for ref in sorted_ref_list:
            n = ClassText.find_index(class_text_list, ref)
            if n > ro.latest_dep:
                ro.latest_dep = n
        return ro

    def is_good(self):
        """Is this class on that should be retained in output."""
        for line in self.lines:
            """
            # for now leave __root__ files since some are still needed
            if line.find('__root__: str') >= 0:
               return False
            # But remove all pass files
            """
            if line.find('pass') >= 0:
                return False
        if not self.name:
            return False
        if self.name.find('min_items') >= 0:
            return False
        # classes like Base64 are ok
        if self.name in numbered_classes:
            return True
        # but any class ending with digit is assumed bad
        if self.name != self.name.rstrip(string.digits):
            return False
        return True


def find_forward_refs(class_list, orders):
    """Find forward references within the file."""
    forward_names = set()
    for c in class_list:
        if c.is_self_ref:
            forward_names.add(c.name)
    for i in range(len(orders)):
        if orders[i].earliest_ref < i:
            forward_names.add(class_list[i].name)

    forward_refs = []
    for c in class_list:
        if c.name in forward_names:
            forward_refs.append(f'{c.name}.update_forward_refs()')
    return forward_refs


def reorder(class_list):
    """Reorder the class list based on the location of its refs and deps."""
    # build list of all class names defined in file
    all_class_names = []
    for c in class_list:
        all_class_names.append(c.name)

    # find direct references for each class in list
    for n, c in enumerate(class_list):
        c.find_direct_refs(all_class_names)
        class_list[n] = c

    # find full references for each class in list with recursion on each ref
    for n, c in enumerate(class_list):
        c.full_refs = c.find_full_refs(class_list)
        class_list[n] = c

    # with full dependency info, now reorder the classes to remove forward refs
    did_swap = True
    loop_num = 0
    orders = None
    while did_swap and loop_num < 1000:
        did_swap = False
        orders = []
        # find the relative placement of each class in list to its references and dependencies
        for c in class_list:
            ro = c.find_order(class_list)
            orders.append(ro)
        # find first class in list out of place and swap its dependency upwards, then break/loop to find new order
        for i, ro in enumerate(orders):
            if ro.latest_dep <= i <= ro.earliest_ref:
                continue
            # pop the out-of-place earliest ref and put it in front
            ct = class_list.pop(ro.earliest_ref)
            class_list.insert(i, ct)
            did_swap = True
            break
        loop_num += 1
    if did_swap:
        logger.info('Excess iteration in reordering!')
    forward_refs = find_forward_refs(class_list, orders)

    # return reordered list of classes with no forward refs
    return class_list, forward_refs


def fix_header(header, needs_conlist):
    """Fix imports and remove timestamp from header."""
    new_header = []
    for r in header:
        # Any is needed for biblio so still need to import it
        # r = re.sub(' Any, ', ' ', r) # noqa: E800
        if r.find(' timestamp: ') >= 0:
            continue
        if needs_conlist:
            # add import of conlist for Union[A, A] case
            r = re.sub(r'^(from\s+pydantic\s+import.*)$', r'\1, conlist', r)
        new_header.append(r)
    return new_header


def good_classes(class_list):
    """Create the list of good classes from the full list."""
    good = []
    for c in class_list:
        if c.is_good():
            good.append(c)
    return good


def clean_classes(class_list):
    """Clean all lines of text in each class for endings that should be removed."""
    for j in range(len(class_list)):
        cls = class_list[j]
        for singleton in singleton_suffixes + singleton_prefixes:
            if cls.name != singleton:
                cls.name = cls.name.replace(singleton, '')
            for i in range(len(cls.lines)):
                line = cls.lines[i]
                if line.find('class ') != 0:
                    if line.find(f'[{singleton}]') == -1:
                        line = line.replace(singleton, '')
                for c2 in class_list:
                    cname = c2.name
                    pattern = cname + '[0-9]*'
                    line = re.sub(pattern, cname, line)
                cls.lines[i] = line
        class_list[j] = cls
    return class_list


def fix_bad_minitems(class_list):
    """Correct behaviour where minitems errors due to optional behaviour."""
    for j in range(len(class_list)):
        cls = class_list[j]
        for i in range(len(cls.lines)):
            line = cls.lines[i]
            neq = line.find('= Field(None')
            if neq > 0 and line.find(cls.name) >= 0:
                cls.lines[i] = line[:neq] + '= None'
            if neq > 0 and line.find('List[Base64]') >= 0:
                cls.lines[i] = line[:neq] + '= None'
        class_list[j] = cls
    return class_list


def fix_file(fname):
    """Fix the Anys in this file and reorder to avoid forward dependencies."""
    all_classes = []
    header = []
    forward_refs = []

    class_text = None
    done_header = False
    needs_conlist = False

    # Find Any's and replace with appropriate singular class
    # Otherwise accumulate all class dependencies for reordering with no forward refs
    with open(fname, 'r') as infile:
        for r in infile.readlines():
            if r.find('.update_forward_refs()') >= 0:
                forward_refs.append(r)
            elif r.find(class_header) == 0:  # start of new class
                done_header = True
                if class_text is not None:  # we are done with current class so add it
                    # prevent anomalous singletons from appearing in output
                    for singleton in singleton_prefixes + singleton_suffixes:
                        if class_text.name.find(singleton) >= 0 and class_text.name != singleton:
                            class_text.name = class_text.name.replace(singleton, '')
                            for i in range(len(class_text.lines)):
                                class_text.lines[i] = class_text.lines[i].replace(singleton, '')
                    all_classes.append(class_text)
                class_text = ClassText(r)
            else:
                if not done_header:  # still in header
                    header.append(r.rstrip())
                else:
                    # fix any line containing Union[A, A] to Union[A, conlist(A, min_items=2)]
                    # this may now include Unions that previously involved anomalous singletons
                    r = re.sub(r'Union\[([^,]*),\s*\1\]', r'Union[\1, conlist(\1, min_items=2)]', r)
                    if r.find(', conlist(') >= 0:
                        needs_conlist = True
                    if r.find('Optional[Dict[str, ') >= 0:
                        match = re.search(r'Optional\[Dict\[str, (.+?)\]\]', r)
                        if match:
                            plural = match.group(1)
                            if plural != 'Any':
                                r = r.replace(plural, plural_lut[plural])
                    for special in special_plurals:
                        if r.find(f'Dict[str, {special}]') >= 0:
                            r = r.replace(special, plural_lut[special])
                    # mark regex strings as raw
                    r = re.sub(r"(\s*regex\s*=\s*)\'(.*)", r"\1r'\2", r)
                    class_text.add_all_refs(r)
                    class_text.add_line(r.rstrip())

    all_classes.append(class_text)  # don't forget final class

    all_classes = good_classes(all_classes)

    all_classes = clean_classes(all_classes)

    all_classes = fix_bad_minitems(all_classes)

    # sort the classes by name in case new versions of DMCG change order of class output
    sorted_classes = sorted(all_classes, key=operator.attrgetter('name'))

    # reorder the classes to remove forward references as much as possible
    # then find any required forward refs and replace the original list with the new ones
    reordered_classes, forward_refs = reorder(sorted_classes)

    header = fix_header(header, needs_conlist)

    # write the classes out in the fixed order
    with open(fname, 'w') as out_file:
        out_file.write('# modified by fix_any.py\n')
        out_file.write(license_header)
        out_file.writelines('\n'.join(header) + '\n')
        classes_written = set()
        for c in reordered_classes:
            if c.name not in classes_written:
                out_file.writelines('\n'.join(c.lines) + '\n')
                classes_written.add(c.name)
        out_file.writelines('\n'.join(forward_refs) + '\n')



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for trestle add command."""
import os
import pathlib
import sys
from unittest.mock import patch

import pytest

from tests import test_utils

import trestle.core.err as err
from trestle.cli import Trestle
from trestle.core.commands.add import AddCmd
from trestle.core.models.actions import UpdateAction
from trestle.core.models.elements import Element, ElementPath
from trestle.core.models.file_content_type import FileContentType
from trestle.oscal.catalog import BackMatter, Catalog
from trestle.utils.fs import get_stripped_contextual_model


def test_add(tmp_path, sample_catalog_minimal):
    """Test AddCmd.add() method for trestle add."""
    # expected catalog after first add of Role
    file_path = pathlib.Path.joinpath(test_utils.JSON_TEST_DATA_PATH, 'minimal_catalog_roles.json')
    expected_catalog_roles1 = Element(Catalog.oscal_read(file_path))

    # expected catalog after second add of Role
    file_path = pathlib.Path.joinpath(test_utils.JSON_TEST_DATA_PATH, 'minimal_catalog_roles_double.json')
    expected_catalog_roles2 = Element(Catalog.oscal_read(file_path))

    # expected catalog after add of Responsible-Party
    file_path = pathlib.Path.joinpath(test_utils.JSON_TEST_DATA_PATH, 'minimal_catalog_roles_double_rp.json')
    expected_catalog_roles2_rp = Element(Catalog.oscal_read(file_path))

    content_type = FileContentType.JSON

    catalog_def_dir, catalog_def_file = test_utils.prepare_trestle_project_dir(
        tmp_path,
        content_type,
        sample_catalog_minimal,
        test_utils.CATALOGS_DIR
    )

    # Execute first _add
    element_path = ElementPath('catalog.metadata.roles')
    catalog_element = Element(sample_catalog_minimal)
    expected_update_action_1 = UpdateAction(expected_catalog_roles1.get_at(element_path), catalog_element, element_path)
    actual_update_action, actual_catalog_roles = AddCmd.add(element_path, Catalog, catalog_element)

    assert actual_catalog_roles == expected_catalog_roles1
    assert actual_update_action == expected_update_action_1

    # Execute second _add - this time roles already exists, so this adds a roles object to roles array
    catalog_element = actual_catalog_roles
    expected_update_action_2 = UpdateAction(expected_catalog_roles2.get_at(element_path), catalog_element, element_path)
    actual_update_action2, actual_catalog_roles2 = AddCmd.add(element_path, Catalog, catalog_element)
    assert actual_catalog_roles2 == expected_catalog_roles2
    assert actual_update_action2 == expected_update_action_2

    # Execute _add for responsible-parties to the same catalog
    element_path = ElementPath('catalog.metadata.responsible-parties')
    catalog_element = actual_catalog_roles2
    expected_update_action_3 = UpdateAction(
        expected_catalog_roles2_rp.get_at(element_path), catalog_element, element_path
    )
    actual_update_action3, actual_catalog_roles2_rp = AddCmd.add(element_path, Catalog, catalog_element)
    assert actual_catalog_roles2_rp == expected_catalog_roles2_rp
    assert actual_update_action3 == expected_update_action_3


def test_add_failure(tmp_path, sample_catalog_minimal):
    """Test AddCmd.add() method for trestle add."""
    content_type = FileContentType.JSON

    catalog_def_dir, catalog_def_file = test_utils.prepare_trestle_project_dir(
        tmp_path,
        content_type,
        sample_catalog_minimal,
        test_utils.CATALOGS_DIR
    )

    element_path = ElementPath('catalog.metadata.*')
    catalog_element = Element(sample_catalog_minimal)

    with pytest.raises(err.TrestleError):
        AddCmd.add(element_path, Catalog, catalog_element)

    element_path = ElementPath('catalog.metadata.title')
    with pytest.raises(err.TrestleError):
        AddCmd.add(element_path, Catalog, catalog_element)

    element_path = ElementPath('catalog.metadata.bad_path')
    with pytest.raises(err.TrestleError):
        AddCmd.add(element_path, Catalog, catalog_element)


def test_run_failure():
    """Test failure of _run for AddCmd."""
    testargs = ['trestle', 'add', '-e', 'catalog.metadata.roles']
    with patch.object(sys, 'argv', testargs):
        with pytest.raises(SystemExit) as e:
            Trestle().run()
        assert e.type == SystemExit
        assert e.value.code == 2

    testargs = ['trestle', 'add', '-f', './catalog.json']
    with patch.object(sys, 'argv', testargs):
        with pytest.raises(SystemExit) as e:
            Trestle().run()
        assert e.type == SystemExit
        assert e.value.code == 2


def test_run(tmp_path, sample_catalog_minimal):
    """Test _run for AddCmd."""
    # expected catalog after add of Responsible-Party
    file_path = pathlib.Path.joinpath(test_utils.JSON_TEST_DATA_PATH, 'minimal_catalog_roles_double_rp.json')
    expected_catalog_roles2_rp = Catalog.oscal_read(file_path)

    content_type = FileContentType.YAML

    catalog_def_dir, catalog_def_file = test_utils.prepare_trestle_project_dir(
        tmp_path,
        content_type,
        sample_catalog_minimal,
        test_utils.CATALOGS_DIR
    )

    testargs = [
        'trestle',
        'add',
        '-f',
        str(catalog_def_file),
        '-e',
        'catalog.metadata.roles, catalog.metadata.roles, catalog.metadata.responsible-parties'
    ]

    with patch.object(sys, 'argv', testargs):
        Trestle().run()

    actual_catalog = Catalog.oscal_read(catalog_def_file)
    assert expected_catalog_roles2_rp == actual_catalog


def test_striped_model(tmp_path, sample_catalog_minimal):
    """Test _run for AddCmd for stripped model."""
    cwd = os.getcwd()
    content_type = FileContentType.JSON
    catalog_def_dir, catalog_def_file = test_utils.prepare_trestle_project_dir(
        tmp_path,
        content_type,
        sample_catalog_minimal,
        test_utils.CATALOGS_DIR
    )
    os.chdir(catalog_def_dir)
    testargs = ['trestle', 'split', '-f', 'catalog.json', '-e', 'catalog.metadata']
    with patch.object(sys, 'argv', testargs):
        Trestle().run()

    # Now that the metadata has been split, add of catalog.metadata.roles will error,
    # but add of catalog.back-matter will pass

    testargs = ['trestle', 'add', '-f', 'catalog.json', '-e', 'catalog.metadata.roles']

    with patch.object(sys, 'argv', testargs):
        assert Trestle().run() == 1

    testargs = ['trestle', 'add', '-f', 'catalog.json', '-e', 'catalog.back-matter']

    current_model, _ = get_stripped_contextual_model()
    current_catalog = current_model.oscal_read(pathlib.Path('catalog.json'))
    current_catalog.back_matter = BackMatter()
    expected_catalog = current_catalog

    with patch.object(sys, 'argv', testargs):
        Trestle().run()

    actual_model, _ = get_stripped_contextual_model()
    actual_catalog = actual_model.oscal_read(pathlib.Path('catalog.json'))
    assert expected_catalog == actual_catalog

    os.chdir(cwd)



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for models util module."""
import pathlib

import pytest

import trestle.core.err as err
import trestle.core.utils as mutils
import trestle.oscal.assessment_plan as assessment_plan
import trestle.oscal.assessment_results as assessment_results
import trestle.oscal.catalog as catalog
import trestle.oscal.component as component
import trestle.oscal.poam as poam
import trestle.oscal.profile as profile
import trestle.oscal.ssp as ssp
import trestle.oscal.target as target


def load_good_catalog() -> catalog.Catalog:
    """Load nist 800-53 as a catalog example."""
    good_sample_path = pathlib.Path('nist-content/nist.gov/SP800-53/rev4/json/NIST_SP-800-53_rev4_catalog.json')

    assert (good_sample_path.exists())
    return catalog.Catalog.oscal_read(good_sample_path)


def test_get_elements() -> None:
    """Test getting flat list of elements."""
    good_sample = load_good_catalog()

    mdlist = mutils.get_elements_of_model_type(good_sample, catalog.Metadata)
    assert (type(mdlist) == list)
    # can only be 1 metadata
    assert (len(mdlist) == 1)
    assert (type(mdlist[0]) == catalog.Metadata)

    control_list = mutils.get_elements_of_model_type(good_sample, catalog.Control)
    assert (len(control_list) >= 1)
    group_list = mutils.get_elements_of_model_type(good_sample, catalog.Group)
    assert (len(group_list) >= 2)


def test_is_collection_field_type() -> None:
    """Test for checking whether the type of a field in an OscalBaseModel object is a collection field."""
    good_catalog = load_good_catalog()

    assert mutils.is_collection_field_type(type('this is a string')) is False

    assert mutils.is_collection_field_type(type(good_catalog)) is False  # Catalog
    catalog_field = catalog.Model.alias_to_field_map()['catalog']
    assert mutils.is_collection_field_type(catalog_field.outer_type_) is False  # Catalog

    assert mutils.is_collection_field_type(type(good_catalog.metadata)) is False  # Metadata
    metadata_field = catalog.Catalog.alias_to_field_map()['metadata']
    assert mutils.is_collection_field_type(metadata_field.outer_type_) is False  # Metadata

    assert mutils.is_collection_field_type(type(good_catalog.metadata.roles)) is False  # list
    roles_field = catalog.Metadata.alias_to_field_map()['roles']
    assert mutils.is_collection_field_type(roles_field.outer_type_) is True  # List[Role]
    assert mutils.is_collection_field_type(roles_field.type_) is False  # Role

    assert mutils.is_collection_field_type(type(good_catalog.metadata.responsible_parties)) is False  # list
    responsible_parties_field = catalog.Metadata.alias_to_field_map()['responsible-parties']
    assert mutils.is_collection_field_type(responsible_parties_field.outer_type_) is True  # Dict[str, ResponsibleParty]
    assert mutils.is_collection_field_type(responsible_parties_field.type_) is False  # ResponsibleParty

    assert mutils.is_collection_field_type(
        type(good_catalog.metadata.parties[0].addresses[0].addr_lines)
    ) is False  # list
    postal_address_field = catalog.Address.alias_to_field_map()['addr-lines']
    assert mutils.is_collection_field_type(postal_address_field.outer_type_) is True  # List[AddrLine]
    assert mutils.is_collection_field_type(postal_address_field.type_) is False  # AddrLine


def test_get_inner_type() -> None:
    """Test retrievel of inner type of a model field representing a collection."""
    good_catalog = load_good_catalog()

    with pytest.raises(err.TrestleError):
        # Type of catalog is not a collection field type
        mutils.get_inner_type(type(good_catalog))

    with pytest.raises(err.TrestleError):
        # Type of field catalog is not a collection field type
        catalog_field = catalog.Model.alias_to_field_map()['catalog']
        mutils.get_inner_type(catalog_field.outer_type_)

    with pytest.raises(err.TrestleError):
        # Type of roles object is not a collection field type
        mutils.get_inner_type(type(good_catalog.metadata.roles))

    # Type of field roles is a collection field type
    roles_field = catalog.Metadata.alias_to_field_map()['roles']
    role_type = mutils.get_inner_type(roles_field.outer_type_)
    assert role_type == catalog.Role

    with pytest.raises(err.TrestleError):
        # Type of responsible_parties object is not a collection field type
        mutils.get_inner_type(type(good_catalog.metadata.responsible_parties))

    # Type of field responsible-parties is a collection field type
    responsible_parties_field = catalog.Metadata.alias_to_field_map()['responsible-parties']
    responsible_party_type = mutils.get_inner_type(responsible_parties_field.outer_type_)
    assert responsible_party_type == catalog.ResponsibleParty


def test_get_root_model() -> None:
    """Test looking for the root model of a trestle oscal module."""
    with pytest.raises(err.TrestleError):
        mutils.get_root_model('invalid')

    with pytest.raises(err.TrestleError):
        mutils.get_root_model('pydantic')

    malias_to_mtype = {
        'catalog': catalog.Catalog,
        'profile': profile.Profile,
        'target-definition': target.TargetDefinition,
        'component-definition': component.ComponentDefinition,
        'system-security-plan': ssp.SystemSecurityPlan,
        'assessment-plan': assessment_plan.AssessmentPlan,
        'assessment-results': assessment_results.AssessmentResults,
        'plan-of-action-and-milestones': poam.PlanOfActionAndMilestones
    }
    for key in malias_to_mtype:
        module_name = malias_to_mtype[key].__module__
        model_type, model_alias = mutils.get_root_model(module_name)
        assert model_type == malias_to_mtype[key]
        assert model_alias == key


def test_classname_to_alias() -> None:
    """Test conversion of class name to alias."""
    module_name = catalog.Catalog.__module__

    with pytest.raises(err.TrestleError):
        mutils.classname_to_alias('any', 'invalid_mode')

    short_classname = catalog.Catalog.__name__
    full_classname = f'{module_name}.{short_classname}'
    json_alias = mutils.classname_to_alias(short_classname, 'json')
    assert json_alias == 'catalog'
    json_alias = mutils.classname_to_alias(full_classname, 'field')
    assert json_alias == 'catalog'

    short_classname = catalog.ResponsibleParty.__name__
    full_classname = f'{module_name}.{short_classname}'
    json_alias = mutils.classname_to_alias(short_classname, 'json')
    assert json_alias == 'responsible-party'
    json_alias = mutils.classname_to_alias(full_classname, 'field')
    assert json_alias == 'responsible_party'

    short_classname = catalog.Property.__name__
    full_classname = f'{module_name}.{short_classname}'
    json_alias = mutils.classname_to_alias(short_classname, 'json')
    assert json_alias == 'property'
    json_alias = mutils.classname_to_alias(full_classname, 'field')
    assert json_alias == 'property'

    short_classname = catalog.MemberOfOrganization.__name__
    full_classname = f'{module_name}.{short_classname}'
    json_alias = mutils.classname_to_alias(short_classname, 'json')
    assert json_alias == 'member-of-organization'
    json_alias = mutils.classname_to_alias(full_classname, 'field')
    assert json_alias == 'member_of_organization'


def test_snake_to_upper_camel() -> None:
    """Ensure Snake to upper camel behaves correctly."""
    cammeled = mutils.snake_to_upper_camel('target_definition')
    assert cammeled == 'TargetDefinition'
    cammeled = mutils.snake_to_upper_camel('control')
    assert cammeled == 'Control'
    cammeled = mutils.snake_to_upper_camel('')
    assert cammeled == ''


def test_camel_to_snake() -> None:
    """Ensure camel to snake behaves correctly."""
    snaked = mutils.camel_to_snake('TargetDefinition')
    assert snaked == 'target_definition'
    snaked = mutils.camel_to_snake('Control')
    assert snaked == 'control'
    snaked = mutils.camel_to_snake('')
    assert snaked == ''


def test_alias_to_classname() -> None:
    """Test alias_to_classname function."""
    assert mutils.alias_to_classname('target-definition', 'json') == 'TargetDefinition'
    assert mutils.alias_to_classname('target_definition', 'field') == 'TargetDefinition'

    with pytest.raises(err.TrestleError):
        assert mutils.alias_to_classname('target-definition', 'invalid') == 'TargetDefinition'


def test_get_target_model() -> None:
    """Test utils method get_target_model."""
    assert mutils.is_collection_field_type(
        mutils.get_target_model(['catalog', 'metadata', 'roles'], catalog.Catalog)
    ) is True
    assert (mutils.get_target_model(['catalog', 'metadata', 'roles'], catalog.Catalog)).__origin__ is list
    assert mutils.get_inner_type(
        mutils.get_target_model(['catalog', 'metadata', 'roles'], catalog.Catalog)
    ) is catalog.Role

    assert mutils.is_collection_field_type(
        mutils.get_target_model(['catalog', 'metadata', 'responsible-parties'], catalog.Catalog)
    ) is True
    assert mutils.get_target_model(['catalog', 'metadata', 'responsible-parties'], catalog.Catalog).__origin__ is dict
    assert mutils.get_inner_type(
        mutils.get_target_model(['catalog', 'metadata', 'responsible-parties'], catalog.Catalog)
    ) is catalog.ResponsibleParty

    assert mutils.is_collection_field_type(
        mutils.get_target_model(['catalog', 'metadata', 'responsible-parties', 'creator'], catalog.Catalog)
    ) is False
    assert mutils.get_target_model(
        ['catalog', 'metadata', 'responsible-parties', 'creator'], catalog.Catalog
    ) is catalog.ResponsibleParty

    assert mutils.get_target_model(['catalog', 'metadata'], catalog.Catalog) is catalog.Metadata

    with pytest.raises(err.TrestleError):
        mutils.get_target_model(['catalog', 'metadata', 'bad_element'], catalog.Catalog)



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for models module."""

import pathlib

from trestle.core import const
from trestle.core import parser

import yaml

yaml_path = pathlib.Path('tests/data/yaml/')
encoding = 'utf8'


def test_parse_dict() -> None:
    """Test parse_dict."""
    file_name = 'good_target.yaml'

    with open(pathlib.Path.joinpath(yaml_path, file_name), 'r', encoding=encoding) as f:
        data = yaml.load(f, yaml.FullLoader)
        target = parser._parse_dict(data['target-definition'], 'trestle.oscal.target.TargetDefinition')
        assert target is not None


def test_to_class_name() -> None:
    """Test to_class_name."""
    tests = [
        {
            'name': 'catalog',
            'expected': 'Catalog',
        },
        {
            'name': 'target-definition',
            'expected': 'TargetDefinition',
        },
    ]

    for test in tests:
        assert parser.to_class_name(test['name']) == test['expected']


def test_to_full_model_name() -> None:
    """Test to_full_model_name."""
    tests = [
        {
            'root_key': 'catalog',
            'name': None,
            'expected': f'{const.PACKAGE_OSCAL}.catalog.Catalog',
        }, {
            'root_key': 'catalog',
            'name': 'group',
            'expected': f'{const.PACKAGE_OSCAL}.catalog.Group',
        }, {
            'root_key': 'target-definition',
            'name': None,
            'expected': f'{const.PACKAGE_OSCAL}.target.TargetDefinition',
        },
        {
            'root_key': 'target',
            'name': 'target-definition',
            'expected': f'{const.PACKAGE_OSCAL}.target.TargetDefinition',
        }, {
            'root_key': 'invalid',
            'name': None,
            'expected': None,
        }
    ]

    for test in tests:
        model_name = parser.to_full_model_name(test['root_key'], test['name'])
        assert model_name == test['expected']


def test_parse_file() -> None:
    """Test parse_file."""
    file_name = 'good_target.yaml'

    tests = [
        {
            'model_name': f'{const.PACKAGE_OSCAL}.target.TargetDefinition', 'expected': 'TargetDefinition'
        }, {
            'model_name': None, 'expected': 'TargetDefinition'
        }
    ]

    for test in tests:
        target = parser.parse_file(pathlib.Path.joinpath(yaml_path, file_name), model_name=test['model_name'])

        assert type(target).__name__ == test['expected']
